/*
 *  Programmazione Grafica 3D - Cap. 6
 *  by Andrea Fasano
 *
 *  Libreria per il loading e la decodifica di un'immagine formato PCX.
 *
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 *
 */

#ifndef _PCX_
#define _PCX_

#include <stdio.h>
#include "demogfx.h"

/****************************************************************************/
/*  COSTANTI                                                                */
/****************************************************************************/
#define PCX_HEADER_SIZE 128

/****************************************************************************/
/*  DEFINIZIONE DEI TIPI                                                    */
/****************************************************************************/
typedef struct {
  char           Manufacturer;
  char           Version;
  char           Encoding;
  char           BitsPerPixel;
  unsigned short Xmin, Ymin, Xmax, Ymax;
  unsigned short HDpi, VDpi;
  char           ColorMap[48];
  char           Reserved;
  char           NumberOfPlanes;
  unsigned short BytesPerLine;
  unsigned short PaletteInfo;
  unsigned short HscreenSize, VscreenSize;
  char           Filler[54];
} PCX_Header;

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
int PCX_XSIZE,    /* Larghezza */
    PCX_YSIZE,    /* Lunghezza */
    PCX_SIZE;     /* Grandezza */

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
VIRTUAL load_pcx (char *filename, TYPE_PALETTE *pal);
void    print_pcx_info (char *filename);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/*-------------------------------------------------------------------------*/
/* FUNZIONE load_pcx                                                       */
/* UTILIZZO Carica e decodifica un file PCX a 256 colori                   */
/* INPUT    filename - Nome del file PCX da caricare                       */
/*          pal      - palette da riempire                                 */
/* OUTPUT   VIRTUAL  - Puntatore all'area di memoria contenente la         */
/*                     immagine decodificata. ATTENZIONE! Tale funzione    */
/*                     alloca la memoria necessaria a contenere l'immagine */
/*                     In caso di un qualsiasi errore, la funzione non     */
/*                     segnala alcun messaggio, ma ritorna solo il valore  */
/*                     NULL                                                */
/*-------------------------------------------------------------------------*/
VIRTUAL load_pcx (char *filename, TYPE_PALETTE *pal)
{
 char           data, num_reps;
 int            count = 0;
 VIRTUAL        image;
 FILE           *in;
 PCX_Header     *pcx_header=(PCX_Header *)malloc(PCX_HEADER_SIZE);

 /* Controlliamo l'esistenza del file, ed eventualmente ritorniamo un errore,
    altrimenti provvediamo a leggere l'header */
 if ((in = fopen (filename, "rb")) == NULL)
  return (NULL);
 else
  fread (pcx_header, PCX_HEADER_SIZE, 1, in);

 /* Controlliamo che esso sia un file PCX, che sia codificato secondo lo RLE,
    e che non sia a 24 bit (cio un'immagine true color) */
 if (pcx_header->Manufacturer != 10) return (NULL);
 if (pcx_header->Encoding != 1) return (NULL);
 if (pcx_header->NumberOfPlanes > 1) return (NULL);

 /* Adesso ci calcoliamo la lunghezza, la larghezza e la dimensione della
    immagine. Essendo globali, le variabili successive potranno tornarci
    utili in numerose occasioni */
 /* invece di calcolare PCX_XSIZE cos:
 PCX_XSIZE = (pcx_header->Xmax - pcx_header->Xmin) + 1;
    usiamo BytesPerLine, in modo da gestire correttamente
    il caso in cui  stato usato un byte di riempimento per avere
    una lunghezza di linea multipla di 2:
 */
 PCX_XSIZE = (pcx_header->BytesPerLine);

 PCX_YSIZE = (pcx_header->Ymax - pcx_header->Ymin) + 1;
 PCX_SIZE  = PCX_XSIZE * PCX_YSIZE;

 /* Allochiamo sufficiente memoria per contenere l'immagine. */
 if ((image = create_screen (PCX_XSIZE, PCX_YSIZE)) == NULL)
   return (NULL);

 /* Ciclo principale di decodifica dell'immagine */
 while (count <= PCX_SIZE)
  {
    /* Leggiamo un byte */
    fread (&data, 1, 1, in);

    /* Controlliamo se il byte letto  un RLE */
    if ((data & 0xc0) == 0xc0) /* S,  un RLE */
      {
       /* Calcoliamo il numero delle ripetizioni */
       num_reps = data & 0x3f;

       /* Leggiamo il colore */
       fread (&data, 1, 1, in);

       /* Memorizziamolo num_reps volte */
       while (num_reps-->0)
         image[count++] = data;
      }
    else
      /* No,  un colore diretto */
      image[count++] = data;
  }

 /* Lettura e normalizzazione della palette */
 fseek (in, -768L, SEEK_END);
 for (count=0; count<256; count++)
  {
   fread (&pal[count].r, 1, 1, in);     pal[count].r = pal[count].r>>2;
   fread (&pal[count].g, 1, 1, in);     pal[count].g = pal[count].g>>2;
   fread (&pal[count].b, 1, 1, in);     pal[count].b = pal[count].b>>2;
  }

 /* Chiudiamo il file */
 fclose (in);

 /* E finalmente l'immagine  pronta per essere usata! */
 return ((VIRTUAL) image);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_pcx_info                                                 */
/* UTILIZZO Stampa a video le caratteristiche generali di un file PCX      */
/* INPUT    filename - Nome del file PCX da caricare                       */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_pcx_info (char *filename)
{
 /*                           o meglio  (sizeof(PCX_Header))    */
 PCX_Header *header=(PCX_Header *)malloc(PCX_HEADER_SIZE);
 FILE       *in;

 /* Controlliamo l'esistenza del file, ed eventualmente ritorniamo un errore,
    altrimenti provvediamo a leggere l'header */
 printf ("\nPCX Info Display");

 if ((in = fopen (filename, "rb")) == NULL)
   {
    printf ("\n* Impossibile trovare il file %s", filename);
    exit (1);
   }
 else
  fread (header, PCX_HEADER_SIZE, 1, in);

 printf ("\n\nFile name: %s\n", filename);

 printf ("\nManufacturer...........");
 (header->Manufacturer == 10)? printf ("ZSoft version") : printf ("Unknown version");

 printf ("\nVersion................");
 switch (header->Version)
   {
     case 0 : printf ("Version 2.5"); break;
     case 2 : printf ("Version 2.8 with palette information"); break;
     case 3 : printf ("Version 2.8 without palette information"); break;
     case 4 : printf ("PcPaintBrush for Windows"); break;
     case 5 : printf ("Version 3.0"); break;
     default : printf ("Unknown version");
   }

 printf ("\nEncoding...............");
 (header->Encoding == 1)? printf ("Run Lenght Encoding") : printf ("Unknown encoding");

 printf ("\nBits per Pixel.........%d", header->BitsPerPixel);

 printf ("\nImage Window...........(%d, %d) - (%d, %d)", header->Xmin, header->Ymin, header->Xmax, header->Ymax);

 printf ("\nHorizontal resolution..%d DPI", header->HDpi);
 printf ("\nVertical resolution....%d DPI", header->VDpi);

 printf ("\nNumber of Planes.......%d", header->NumberOfPlanes);

 printf ("\nBytes per Line.........%d", header->BytesPerLine);

 printf ("\nPalette Info...........");
 switch (header->PaletteInfo)
   {
     case 1 : printf ("Color or Black and White"); break;
     case 2 : printf ("Grayscale"); break;
     default : printf ("Unknown");
   }

 putchar('\n');

 fclose (in);
}

#endif
