/*
 *  Programmazione Grafica 3D - Cap. 8
 *  by Andrea Fasano
 *
 *  Libreria per il loading e la decodifica di un'immagine formato PCX.
 *
 *
 *  convertito per DJGPP da Antonio Bonifati
 *  website: http://Assembly_Page_di_Ant.freeweb.org
 */

#ifndef _PCX_
#define _PCX_

#include <stdio.h>
#include "demogfx.h"

/****************************************************************************/
/*  COSTANTI                                                                */
/****************************************************************************/
#define PCX_HEADER_SIZE 128

/****************************************************************************/
/*  DEFINIZIONE DEI TIPI                                                    */
/****************************************************************************/
typedef struct {
  char           Manufacturer;
  char           Version;
  char           Encoding;
  char           BitsPerPixel;
  unsigned short Xmin, Ymin, Xmax, Ymax;
  unsigned short HDpi, VDpi;
  char           ColorMap[48];
  char           Reserved;
  char           NumberOfPlanes;
  unsigned short BytesPerLine;
  unsigned short PaletteInfo;
  unsigned short HscreenSize, VscreenSize;
  char           Filler[54];
} PCX_Header;

/****************************************************************************/
/*  VARIABILI GLOBALI                                                       */
/****************************************************************************/
int PCX_XSIZE,    /* Larghezza */
    PCX_YSIZE,    /* Lunghezza */
    PCX_SIZE;     /* Grandezza */
char pcx_row[640*4]; /** Array di comodo **/

/****************************************************************************/
/*  PROTOTIPI                                                               */
/****************************************************************************/
VIRTUAL load_pcx (char *filename, TYPE_PALETTE *pal);
void    print_pcx_info (char *filename);
void    read_row (FILE *in, int row_lenght);
VIRTUAL load_pcx_truecolor (char *filename);

/****************************************************************************/
/*  FUNZIONI                                                                */
/****************************************************************************/
/*-------------------------------------------------------------------------*/
/* FUNZIONE load_pcx                                                       */
/* UTILIZZO Carica e decodifica un file PCX a 256 colori                   */
/* INPUT    filename - Nome del file PCX da caricare                       */
/*          pal      - palette da riempire                                 */
/* OUTPUT   VIRTUAL  - Puntatore all'area di memoria contenente la         */
/*                     immagine decodificata. ATTENZIONE! Tale funzione    */
/*                     alloca la memoria necessaria a contenere l'immagine */
/*                     In caso di un qualsiasi errore, la funzione non     */
/*                     segnala alcun messaggio, ma ritorna solo il valore  */
/*                     NULL                                                */
/*-------------------------------------------------------------------------*/
VIRTUAL load_pcx (char *filename, TYPE_PALETTE *pal)
{
 char           data, num_reps;
 int            count = 0;
 VIRTUAL        image;
 FILE           *in;
 PCX_Header     *pcx_header=(PCX_Header *)malloc(PCX_HEADER_SIZE);

 /* Controlliamo l'esistenza del file, ed eventualmente ritorniamo un errore,
    altrimenti provvediamo a leggere l'header */
 if ((in = fopen (filename, "rb")) == NULL)
  return (NULL);
 else
  fread (&pcx_header, PCX_HEADER_SIZE, 1, in);

 /* Controlliamo che esso sia un file PCX, che sia codificato secondo lo RLE,
    e che non sia a 24 bit (cio un'immagine true color) */
 if (pcx_header->Manufacturer != 10) return (NULL);
 if (pcx_header->Encoding != 1) return (NULL);
 if (pcx_header->NumberOfPlanes > 1) return (NULL);

 /* Adesso ci calcoliamo la lunghezza, la larghezza e la dimensione della
    immagine. Essendo globali, le variabili successive potranno tornarci
    utili in numerose occasioni */
 PCX_XSIZE = (pcx_header->Xmax - pcx_header->Xmin) + 1;
 PCX_YSIZE = (pcx_header->Ymax - pcx_header->Ymin) + 1;
 PCX_SIZE  = PCX_XSIZE * PCX_YSIZE;

 /* Allochiamo sufficiente memoria per contenere l'immagine. */
 if ((image = create_screen (PCX_XSIZE, PCX_YSIZE)) == NULL)
   return (NULL);

 /* Ciclo principale di decodifica dell'immagine */
 while (count <= PCX_SIZE)
  {
    /* Leggiamo un byte */
    fread (&data, 1, 1, in);

    /* Controlliamo se il byte letto  un RLE */
    if ((data & 0xc0) == 0xc0) /* S,  un RLE */
      {
       /* Calcoliamo il numero delle ripetizioni */
       num_reps = data & 0x3f;

       /* Leggiamo il colore */
       fread (&data, 1, 1, in);

       /* Memorizziamolo num_reps volte */
       while (num_reps-->0)
         image[count++] = data;
      }
    else
      /* No,  un colore diretto */
      image[count++] = data;
  }

 /* Lettura e normalizzazione della palette */
 fseek (in, -768L, SEEK_END);
 for (count=0; count<256; count++)
  {
   fread (&pal[count].r, 1, 1, in);     pal[count].r = pal[count].r>>2;
   fread (&pal[count].g, 1, 1, in);     pal[count].g = pal[count].g>>2;
   fread (&pal[count].b, 1, 1, in);     pal[count].b = pal[count].b>>2;
  }

 /* Chiudiamo il file */
 fclose (in);

 /* liberiamo la memoria usata per l'intestazione */
 /* (potevamo usare anche una locale)             */
 free(pcx_header);

 /* E finalmente l'immagine  pronta per essere usata! */
 return ((VIRTUAL) image);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE print_pcx_info                                                 */
/* UTILIZZO Stampa a video le caratteristiche generali di un file PCX      */
/* INPUT    filename - Nome del file PCX da caricare                       */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void print_pcx_info (char *filename)
{
 PCX_Header *header=(PCX_Header *)malloc(PCX_HEADER_SIZE);
 FILE       *in;

 /* Controlliamo l'esistenza del file, ed eventualmente ritorniamo un errore,
    altrimenti provvediamo a leggere l'header */
 printf ("\nPCX Info Display");
 if ((in = fopen (filename, "rb")) == NULL)
   {
    printf ("\n* Impossibile trovare il file %s", filename);
    exit (0);
   }
 else
  fread (header, PCX_HEADER_SIZE, 1, in);

 printf ("\n\nFile name: %s\n", filename);

 printf ("\nManufacturer...........");
 (header->Manufacturer == 10)? printf ("ZSoft version") : printf ("Unknown version");

 printf ("\nVersion................");
 switch (header->Version)
   {
     case 0 : printf ("Version 2.5"); break;
     case 2 : printf ("Version 2.8 with palette information"); break;
     case 3 : printf ("Version 2.8 without palette information"); break;
     case 4 : printf ("PcPaintBrush for Windows"); break;
     case 5 : printf ("Version 3.0"); break;
     default : printf ("Unknown version");
   }

 printf ("\nEncoding...............");
 (header->Encoding == 1)? printf ("Run Lenght Encoding") : printf ("Unknown encoding");

 printf ("\nBits per Pixel.........%d", header->BitsPerPixel);

 printf ("\nImage Window...........(%d, %d) - (%d, %d)", header->Xmin, header->Ymin, header->Xmax, header->Ymax);

 printf ("\nHorizontal resolution..%d DPI", header->HDpi);
 printf ("\nVertical resolution....%d DPI", header->VDpi);

 printf ("\nNumber of Planes.......%d", header->NumberOfPlanes);

 printf ("\nBytes per Line.........%d", header->BytesPerLine);

 printf ("\nPalette Info...........");
 switch (header->PaletteInfo)
   {
     case 1 : printf ("Color or Black and White"); break;
     case 2 : printf ("Grayscale"); break;
     default : printf ("Unknown");
   }

 putchar('\n');

 free(header);

 fclose (in);
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE read_row_truecolor                                             */
/* UTILIZZO Legge e decodifica una scanline di un file a 16 milioni di     */
/*          colori. E' utilizzata dalla funzione load_pcx_truecolor        */
/* INPUT    in - Nome del file PCX                                         */
/*          row_lenght - lunghezza in byte di un piano                     */
/* OUTPUT   Nessuno                                                        */
/*-------------------------------------------------------------------------*/
void read_row_truecolor (FILE *in, int row_lenght)
{
 unsigned short num_bytes, len=0;
 unsigned char data;

 while (len < row_lenght)
  {
    if (fread (&data, 1, 1, in) == -1)
     {
       set_text_mode ();
       printf ("\nErrore di lettura!");
       exit(1);
     }

    if (data >= 192 && data <=255)
      {
       num_bytes = data - 192;
       fread (&data, 1, 1, in);

       while (num_bytes-->0)
         pcx_row[len++] = (unsigned char) data; /* cast superfluo:        */
      }                                         /* lo metto per chiarezza */
    else pcx_row[len++] = (unsigned char) data;
   }
}

/*-------------------------------------------------------------------------*/
/* FUNZIONE load_pcx_truecolor                                             */
/* UTILIZZO Carica e decodifica un file PCX a 16 milioni di colori         */
/* INPUT    filename - Nome del file PCX da caricare                       */
/* OUTPUT   VIRTUAL  - Puntatore all'area di memoria contenente la         */
/*                     immagine decodificata. ATTENZIONE! Tale funzione    */
/*                     alloca la memoria necessaria a contenere l'immagine */
/*                     In caso di un qualsiasi errore, la funzione non     */
/*                     segnala alcun messaggio, ma ritorna solo il valore  */
/*                     NULL                                                */
/* NB: Si presuppone di attivare una modalit grafica a 32 bit             */
/*-------------------------------------------------------------------------*/
VIRTUAL load_pcx_truecolor (char *filename)
{
 int            x, y, lun;
 VIRTUAL        image;
 FILE           *in;
 PCX_Header     *pcx_header=(PCX_Header *)malloc(PCX_HEADER_SIZE);

 if ((in = fopen (filename, "rb")) == NULL)
  return (NULL);
 else
  fread (pcx_header, PCX_HEADER_SIZE, 1, in);

 PCX_XSIZE = (pcx_header->Xmax - pcx_header->Xmin) + 1;
 PCX_YSIZE = (pcx_header->Ymax - pcx_header->Ymin) + 1;
 PCX_SIZE  = PCX_XSIZE * PCX_YSIZE * (pcx_header->NumberOfPlanes);

 /*** L'immagine  a 32 bit, mentre il file .pcx  a 24 bit ***/
 image = (VIRTUAL) malloc (PCX_XSIZE * PCX_YSIZE * (pcx_header->NumberOfPlanes+1));

 /*** Inizio ciclo di decodifica dell'immagine ***/
 for (y=0; y<PCX_YSIZE*4; y+=4)
 {
   read_row_truecolor (in, PCX_XSIZE*3);

   lun=0;
   for (x=0; x<PCX_XSIZE*4; x+=4)
    {
      image[x+y*PCX_XSIZE]   = (char) pcx_row [lun+PCX_XSIZE*2];
      image[x+1+y*PCX_XSIZE] = (char) pcx_row [lun+PCX_XSIZE];
      image[x+2+y*PCX_XSIZE] = (char) pcx_row [lun+0];
      image[x+3+y*PCX_XSIZE] = (char) 0;
      lun++;
    }
  }

 fclose (in);

 /* liberiamo la memoria usata per l'intestazione */
 /* (potevamo usare anche una locale)             */
 free(pcx_header);

 return ((VIRTUAL) image);
}


#endif
